import json

import numpy as np

from ..argtypes import list_of
from ..genome import ModelGenome


class BaseOptimizer:
    def __init__(self, config, model):
        self.config = config
        self.model = model
        self.rng = np.random.RandomState(self.config.seed)

    def run(self, f_run_episode):
        num_episodes = 0
        self.pre_run()
        while True:
            print('Episode {}'.format(num_episodes))
            self.pre_episode(num_episodes)
            reward, steps = f_run_episode(self.genome)
            self.post_episode(num_episodes, reward, steps)
            print('Reward {} in {} steps'.format(reward, steps))
            num_episodes += 1

    def pre_run(self):
        pass

    def pre_episode(self, episode):
        pass

    def post_episode(self, episode, reward, steps):
        pass

    @classmethod
    def add_config_to_parser(cls, parser):
        parser.add_argument('--gene-weight-ratio', type=float, default=0.001)
        parser.add_argument('--freq-weight-ratio', type=float, default=1.)
        parser.add_argument('--i-sigma', type=float, default=1.)
        parser.add_argument('--v-sigma', type=list_of(float), default=1.)
        parser.add_argument('--v-init', type=list_of(float), default=(-1, 1.))
        parser.add_argument('--seed', type=int, default=int('beef', 16))
        parser.add_argument('--clear-store', action='store_true')
        parser.add_argument('--redis-params', type=json.loads, default={'host': 'localhost'})



import hashlib

import numpy as np
import redis
import scipy.stats


class GenePool:
    def __init__(self, key='genescores', redis_params=None):
        self.key = key
        if redis_params is None:
            redis_params = dict(host='localhost')
        self.redis = redis.StrictRedis(**redis_params)

    def report_score(self, genome, score):
        '''Saves a history of each genome's fitness and saves the mean in
        a sorted set'''
        data = genome.serialize_genomes()
        genome_key = self.genome_key(data)
        self.redis.rpush(genome_key, score)
        history = self.redis.lrange(genome_key, 0, -1)
        history = list(map(float, history))
        self.redis.zadd(self.key, np.mean(history), data)

    def top_n(self, n, reverse=True):
        if reverse:
            f = self.redis.zrevrange
        else:
            f = self.redis.zrange
        results = f(self.key, 0, n - 1, withscores=True)
        return results

    def clear(self):
        self.redis.flushdb()

    def genome_key(self, serialized_genome):
        return '{}/gene/{}'.format(self.key, hashlib.sha1(serialized_genome).hexdigest())



import json

import numpy as np

from ..argtypes import list_of
from ..genome import ModelGenome
from ..scoreboard import AgentScoreboard
from .base import BaseOptimizer


class Optimizer(BaseOptimizer):
    '''"Asynchronous Gene Pool"

    This optimizer evaluates genome fitness in parallel and asynchronously.
    The gene pool--a set of all genomes in the popuation ordered by their
    fitness--is maintained by a group of workers which pull highly-fit genomes
    and report the fitness of their mutant offspring.
    '''
    def __init__(self, config, model, agent, genepool):
        super(Optimizer, self).__init__(config, model)
        self.agent = agent
        self.genepool = genepool

    def post_episode(self, episode, reward, num_steps):
        self.genepool.report_score(self.agent.genome, reward)
        best_genomes = self.genepool.top_n(self.config.num_best)
        _, best_score = best_genomes[0]
        _, worst_best_score = best_genomes[-1]
        print('Genepool top: {}, {}'.format(best_score, worst_best_score))
        if best_genomes and self.rng.uniform() < 0.1:
            # replay top genomes to make sure they're not flukes
            best_genome, _ = best_genomes[self.rng.randint(len(best_genomes))]
            self.agent.load_genome(best_genome)
        else:
            if reward < 0.5 * (worst_best_score + best_score) and len(best_genomes) > 1:
                self.agent.crossover(best_genomes)
            self.agent.mutate(index_sigma=self.config.i_sigma, value_sigma=self.config.v_sigma)
        self.agent.update_model()

    @property
    def genome(self):
        return self.agent.genome

    @classmethod
    def add_config_to_parser(cls, parser):
        super(Optimizer, cls).add_config_to_parser(parser)
        parser.add_argument('--num-best', type=int, default=30)
        parser.add_argument('--num-agents', type=int, default=10)



#!/usr/bin/env python
from distutils.core import setup
from setuptools import find_packages

setup(
    name='pytorch-cns',
    version='0.2.2',
    description='Generalized Compressed Network Search with PyTorch',
    author='Adam Wentz',
    author_email='adam@adamwentz.com',
    url='https://github.com/awentzonline/pytorch-cns',
    packages=find_packages(),
    install_requires=[
        'numpy',
        'torch',
        'torchvision',
        'scipy',
        'six',
        'redis',
    ]
)
